import collections
import functools
from os import PathLike
import pathlib
import sys
from types import CodeType, FrameType, ModuleType
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Dict,
    Set,
    Tuple,
    Union,
)
import warnings

from .call_capture import CallCapture, CallListCapture
from .common import Handle, TraceFunc, suspend, resume, is_suspended, THREAD_LOCAL
from .scope import Scope
from .breakpoint import Breakpoint


__all__ = [
    "NoPdb",
    "breakpoint",
    "capture_call",
    "capture_calls",
    "get_nopdb",
]


class NoPdb:
    """The main NoPdb class.

    Multiple instances can be created, but only one can be active in a given thread at
    a given time. It can be used as a context manager.
    """

    def __init__(self):
        self._started = False
        self._orig_trace_func: Optional[TraceFunc] = None
        self._callbacks: Dict[
            Handle, Tuple[Scope, Set[str], TraceFunc]
        ] = collections.OrderedDict()

    @property
    def started(self) -> bool:
        return self._started

    def start(self) -> None:
        """Start this instance.

        Called automatically when the object is used as a context manager.
        """
        if self._started:
            raise RuntimeError("nopdb has already been started")
        self._orig_trace_func = sys.gettrace()
        sys.settrace(self._trace_func)
        self._started = True

    def stop(self) -> None:
        """Stop this instance.

        Called automatically when the object is used as a context manager.
        """
        if not self._started:
            raise RuntimeError("nopdb has not been started")
        if getattr(sys.gettrace(), "__self__", None) is self:
            sys.settrace(self._orig_trace_func)
        else:
            warnings.warn(
                "Another trace function has been set since nopdb was started. "
                "Will not restore the original trace function.",
                RuntimeWarning,
            )
        self._started = False

    def _ensure_started(self):
        started = self._started
        if started and getattr(sys.gettrace(), "__self__", None) is not self:
            raise RuntimeError(
                "nopdb has been started, but a different trace function was "
                "set in the meantime"
            )
        if not started:
            self.start()

    def __enter__(self) -> "NoPdb":
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.stop()

    def _trace_func(
        self, frame: FrameType, event: str, arg: Any
    ) -> Optional[TraceFunc]:
        # Do not do anything if this instance is not currently tracing
        if getattr(sys.gettrace(), "__self__", None) is not self:
            return None
        if is_suspended():
            return None

        # Avoid tracing our own code
        try:
            # Replacement for PurePath.is_relative_to() for Python < 3.9
            pathlib.PurePath(frame.f_code.co_filename).relative_to(_NOPDB_ROOT)
            return None
        except ValueError:
            pass

        trace_locally = False

        # Check which callbacks we need to call
        for scope, events, callback in self._callbacks.values():
            if scope.match_frame(frame):
                # If an event other than 'call' is requested, we will need to
                # return a local trace function.
                if event == "call" and not all(e == "call" for e in events):
                    trace_locally = True

                if event in events:
                    callback(frame, event, arg)

        if trace_locally:
            return self._trace_func

        return None

    def add_callback(
        self, scope: Scope, callback: TraceFunc, events: Iterable[str]
    ) -> Handle:
        """Register a low-level callback for the given type(s) of events.

        Args:
            scope (Scope): The scope in which the callback should be active.
            callback (TraceFunc): The callback function. It should have the same
                signature as the function passed to :func:`sys.settrace`, but its
                return value will be ignored.
            events (~typing.Iterable[str]): A list of event names (:code:`'call'`,
                :code:`'line'`, :code:`'return'` or :code:`'exception'`); see
                :func:`sys.settrace`.

        Returns:
            Handle:
                A handle that can be passed to :meth:`remove_callback`.
        """
        handle = Handle()
        self._callbacks[handle] = (scope, set(events), callback)
        return handle

    def remove_callback(self, handle: Handle) -> None:
        """Remove a callback added using :meth:`add_callback`.

        Args:
            handle (Handle): A handle returned by :meth:`add_callback`.
        """
        del self._callbacks[handle]

    def capture_call(
        self,
        function: Optional[Union[Callable, str]] = None,
        *,
        module: Optional[ModuleType] = None,
        file: Optional[Union[str, PathLike]] = None,
        unwrap: bool = True
    ) -> CallCapture:
        """Capture a function call.

        The returned object can be used as a context manager, which will cause the
        capturing to stop at the end of the block.

        If multiple calls occur, the returned object will be updated as each call
        returns. At the end, the returned object will contain information about the
        call that was the last to return.

        Args:
            function (~typing.Callable or str, optional): A Python callable or the name
                of a Python function. If an instance method is passed, only calls
                invoked on that particular instance will be captured.
            module (~types.ModuleType, optional): A Python module. If given, only calls
                to functions defined in this module will be captured.
            file (str or ~os.PathLike, optional): A path to a Python source file. If
                given, only calls to functions defined in this file will be captured.
                If a string is passed, it will be used as a glob-style pattern for
                :meth:`pathlib.PurePath.match`. If a path-like object is passed, it
                will be resolved to a canonical path and checked for an exact match.
            unwrap (bool, optional): Whether or not to unwrap `function` when it is a
                wrapper (e.g. produced by a decorator). Only works when `function` is
                given as a callable. Defaults to `True`.

        Returns:
            CallCapture:
                An instance of :class:`CallInfo` which also works as a context
                manager.
        """
        suspend()
        try:
            return CallCapture(
                nopdb=self, scope=Scope(function, module, file, unwrap=unwrap)
            )
        finally:
            resume()

    def capture_calls(
        self,
        function: Optional[Union[Callable, str]] = None,
        *,
        module: Optional[ModuleType] = None,
        file: Optional[Union[str, PathLike]] = None,
        unwrap: bool = True
    ) -> CallListCapture:
        """Capture function calls.

        The return value is an initially empty list, which is updated with a new item
        as each call returns. At the end, the list will contain a :class:`CallInfo`
        object for each call, following the order in which the calls returned.

        The return value can also be used as a context manager, which will cause the
        capturing to stop at the end of the block.

        Args:
            function (~typing.Callable or str, optional): A Python callable or the name
                of a Python function. If an instance method is passed, only calls
                invoked on that particular instance will be captured.
            module (~types.ModuleType, optional): A Python module. If given, only calls
                to functions defined in this module will be captured.
            file (str or ~os.PathLike, optional): A path to a Python source file. If
                given, only calls to functions defined in this file will be captured.
                If a string is passed, it will be used as a glob-style pattern for
                :meth:`pathlib.PurePath.match`. If a path-like object is passed, it
                will be resolved to a canonical path and checked for an exact match.
            unwrap (bool, optional): Whether or not to unwrap `function` when it is a
                wrapper (e.g. produced by a decorator). Only works when `function` is
                given as a callable. Defaults to `True`.

        Returns:
            CallListCapture:
                A list of :class:`CallInfo` objects which also works as a
                context manager.
        """
        suspend()
        try:
            return CallListCapture(
                nopdb=self, scope=Scope(function, module, file, unwrap=unwrap)
            )
        finally:
            resume()

    def breakpoint(
        self,
        function: Optional[Union[Callable, str]] = None,
        *,
        module: Optional[ModuleType] = None,
        file: Optional[Union[str, PathLike]] = None,
        line: Optional[Union[int, str]] = None,
        cond: Optional[Union[str, bytes, CodeType]] = None,
        unwrap: bool = True
    ) -> Breakpoint:
        """Set a breakpoint.

        The returned :class:`Breakpoint` object works as a context manager that removes
        the breakpoint at the end of the block.

        The breakpoint itself does not stop execution when hit, but can trigger
        user-defined actions; see :meth:`Breakpoint.eval`, :meth:`Breakpoint.exec`,
        :meth:`Breakpoint.debug`.

        At least a function, a module or a file must be specified. If no function is
        given, a line is also required.

        Example::

           # Stop at the line in f that says "return y"
           with nopdb.breakpoint(function=f, line="return y") as bp:
               x = bp.eval("x")             # Schedule an expression
               type_y = bp.eval("type(y)")  # Another one
               bp.exec("print(y)")          # Schedule a print statement

               # Now run some code that calls f
               # ...

           print(x, type_y)  # Retrieve the recorded values

        Args:
            function (~typing.Callable or str, optional): A Python callable or the name
                of a Python function. If an instance method is passed, only calls
                invoked on that particular instance will trigger the breakpoint.
            module (~types.ModuleType, optional): A Python module.
            file (str or ~os.PathLike, optional): A path to a Python source file.
                If a string is passed, it will be used as a glob-style pattern for
                :meth:`pathlib.PurePath.match`. If a path-like object is passed, it
                will be resolved to a canonical path and checked for an exact match.
            line (int or str, optional): The line at which to break. Either of the
                following:

                * The line number, counted from the beginning of the file.
                * The source code of the line. The code needs to match exactly, except
                  for leading and trailing whitespace.
                * `None`; in this case, a `function` must be passed, and the breakpoint
                  will be triggered every time the function is called.

                Note that unlike in `pdb`, the breakpoint will only get triggered by
                the exact given line. This means that some lines will not work as
                breakpoints, e.g. if they are part of a multiline statement or do not
                contain any code to execute.
            cond (str, bytes or ~types.CodeType, optional): A condition to evaluate. If
                given, the breakpoint will only be triggered when the condition
                evaluates to true.
            unwrap (bool, optional): Whether or not to unwrap `function` when it is a
                wrapper (e.g. produced by a decorator). Only works when `function` is
                given as a callable. Defaults to `True`.

        Returns:
            Breakpoint:
                The breakpoint object, which also works as a context manager.
        """
        suspend()
        try:
            return Breakpoint(
                nopdb=self,
                scope=Scope(function, module, file, unwrap=unwrap),
                line=line,
                cond=cond,
            )
        finally:
            resume()


def get_nopdb() -> NoPdb:
    """Return an instance of :class:`NoPdb`.

    If a :class:`NoPdb` instance is currently active in the current thread, that
    instance is returned. Otherwise, the default instance for the current thread is
    returned.
    """
    # If a NoPdb instance is currently tracing, return it
    trace_obj = getattr(sys.gettrace(), "__self__", None)
    if isinstance(trace_obj, NoPdb):
        return trace_obj

    # Otherwise return the default instance for this thread
    if not hasattr(THREAD_LOCAL, "default_nopdb"):
        THREAD_LOCAL.default_nopdb = NoPdb()
    return THREAD_LOCAL.default_nopdb


@functools.wraps(get_nopdb().capture_call)
def capture_call(*args, **kwargs):
    return get_nopdb().capture_call(*args, **kwargs)


@functools.wraps(get_nopdb().capture_calls)
def capture_calls(*args, **kwargs):
    return get_nopdb().capture_calls(*args, **kwargs)


@functools.wraps(get_nopdb().breakpoint)
def breakpoint(*args, **kwargs):
    return get_nopdb().breakpoint(*args, **kwargs)


_NOPDB_ROOT = pathlib.PurePath(__file__).parent



import abc
import collections
import inspect
import traceback
from types import FrameType
from typing import Any, List, Optional, Dict, TYPE_CHECKING

from .common import NoPdbContextManager

if TYPE_CHECKING:
    from .nopdb import NoPdb
    from .scope import Scope


class CallInfo:
    """Information about a function call.

    Attributes:
        name (str): The name of the function's code object.
        file (str): The path to the file where the function was defined.
        stack (traceback.StackSummary): The call stack.
        args (dict): The function's arguments.
        locals (dict): Local variables on return.
        globals (dict): Global variables on return.
        return_value: The return value.
    """

    def __init__(self):
        self.name: Optional[str] = None
        self.file: Optional[str] = None
        self.stack: Optional[traceback.StackSummary] = None
        self.args: Optional[dict] = None
        self.locals: Optional[dict] = None
        self.globals: Optional[dict] = None
        self.return_value: Optional[Any] = None

    def __repr__(self) -> str:
        return "{}(name={!r}, args={}({}), return_value={!r})".format(
            type(self).__name__,
            self.name,
            type(self.args).__name__,
            ", ".join("{}={!r}".format(k, v) for k, v in (self.args or {}).items()),
            self.return_value,
        )

    def print_stack(self, file=None) -> None:
        """Print the call stack."""
        if self.stack is not None:
            for line in self.stack.format():
                print(line, end="", file=file)


class BaseCallCapture:
    def __init__(self):
        self._result_by_frame: Dict[FrameType, CallInfo] = {}

    def _callback(self, frame: FrameType, event: str, arg: Any):
        if frame not in self._result_by_frame:
            self._result_by_frame[frame] = CallInfo()
        result = self._result_by_frame[frame]

        if event == "call":
            result.name = frame.f_code.co_name
            result.file = frame.f_code.co_filename
            result.stack = traceback.extract_stack(frame)
            result.return_value = None

            # Extract argument values
            arg_info = inspect.getargvalues(frame)
            result.args = collections.OrderedDict(
                [
                    (name, arg_info.locals[name])
                    for name in [*arg_info.args, arg_info.varargs, arg_info.keywords]
                    if name in arg_info.locals
                ]
            )
        elif event == "return":
            result.locals = frame.f_locals
            result.globals = frame.f_globals
            result.return_value = arg

            # This frame is done, save the result
            del self._result_by_frame[frame]
            self._update_result(result)

    @abc.abstractmethod
    def _update_result(self, result: CallInfo) -> None:
        pass


class CallCapture(BaseCallCapture, NoPdbContextManager, CallInfo):
    def __init__(self, nopdb: "NoPdb", scope: "Scope"):
        BaseCallCapture.__init__(self)
        NoPdbContextManager.__init__(
            self, nopdb=nopdb, scope=scope, events=["call", "return"]
        )
        CallInfo.__init__(self)

    def _update_result(self, result: CallInfo) -> None:
        self.__dict__.update(result.__dict__)


class CallListCapture(BaseCallCapture, NoPdbContextManager, List[CallInfo]):
    def __init__(self, nopdb: "NoPdb", scope: "Scope"):
        BaseCallCapture.__init__(self)
        NoPdbContextManager.__init__(
            self, nopdb=nopdb, scope=scope, events=["call", "return"]
        )
        list.__init__(self)

    def _update_result(self, result: CallInfo) -> None:
        self.append(result)



import bdb
import sys
from types import FrameType
from typing import Any, Callable, Dict, Optional, Tuple, Type

from .common import TraceFunc


def get_nice_debugger(frame: FrameType, cls: Type[bdb.Bdb], kwargs: dict):
    class NiceDebugger(cls):  # type: ignore
        def __init__(self, frame: FrameType, kwargs: dict):
            self._orig_trace_func = sys.gettrace()
            self._orig_local_trace_funcs = _get_local_trace_funcs(frame)
            super().__init__(**kwargs)

        def set_continue(self) -> None:
            super().set_continue()
            if sys.gettrace() is None:
                self._restore_trace_funcs()

        def trace_dispatch(
            self, frame: FrameType, event: str, arg: Any
        ) -> Optional[Callable[[FrameType, str, Any], Any]]:
            if id(frame) not in self._orig_local_trace_funcs:
                self._orig_local_trace_funcs[id(frame)] = (frame, frame.f_trace)

            try:
                return super().trace_dispatch(frame, event, arg)
            except bdb.BdbQuit:
                self._restore_trace_funcs()
                return None

        def _restore_trace_funcs(self) -> None:
            _restore_local_trace_funcs(self._orig_local_trace_funcs)
            sys.settrace(self._orig_trace_func)

    return NiceDebugger(frame, kwargs)


def _get_local_trace_funcs(
    frame: FrameType,
) -> Dict[int, Tuple[FrameType, Optional[TraceFunc]]]:
    trace_funcs = {}
    while True:
        # Store a reference to the frame to prevent its ID from being reused by another
        # frame
        trace_funcs[id(frame)] = (frame, frame.f_trace)
        if not frame.f_back:
            break
        frame = frame.f_back
    return trace_funcs


def _restore_local_trace_funcs(
    local_trace_funcs: Dict[int, Tuple[FrameType, Optional[TraceFunc]]]
) -> None:
    for frame, trace_func in local_trace_funcs.values():
        if trace_func is None:
            del frame.f_trace
        else:
            frame.f_trace = trace_func



import abc
import threading
from types import FrameType
from typing import Any, Callable, List, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from .nopdb import NoPdb
    from .scope import Scope


TraceFunc = Callable[[FrameType, str, Any], Any]


THREAD_LOCAL = threading.local()


class Handle:
    pass


# Some functions for suspending and resuming tracing internally. We cannot have
# a @contextlib.contextmanager for this, because the __enter__ would end up being
# traced. So we just define two functions and use them with try-finally.
# Note that these are only necessary to use in code that uses other libraries.


def suspend():
    THREAD_LOCAL.suspended = getattr(THREAD_LOCAL, "suspended", 0) + 1


def resume():
    THREAD_LOCAL.suspended -= 1


def is_suspended():
    return getattr(THREAD_LOCAL, "suspended", 0) > 0


TNoPdbContextManager = TypeVar("TNoPdbContextManager", bound="NoPdbContextManager")


class NoPdbContextManager(abc.ABC):
    def __init__(self, nopdb: "NoPdb", scope: "Scope", events: List[str]):
        self._nopdb = nopdb
        self._scope = scope
        self._events = events

        self._was_started = False
        self._handle = nopdb.add_callback(self._scope, self._callback, self._events)

    @abc.abstractmethod
    def _callback(self, frame: FrameType, event: str, arg: Any) -> None:
        pass

    def __enter__(self: TNoPdbContextManager) -> TNoPdbContextManager:
        self._was_started = self._nopdb.started
        self._nopdb._ensure_started()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        try:
            self._nopdb.remove_callback(self._handle)
        finally:
            if self._nopdb.started and not self._was_started:
                self._nopdb.stop()



import bdb
import functools
import linecache
import pdb
import sys
from types import CodeType, FrameType
from typing import Any, Callable, List, Optional, Dict, Type, Union, TYPE_CHECKING
import warnings

from .common import NoPdbContextManager
from .nice_debugger import get_nice_debugger
from .scope import Scope

if TYPE_CHECKING:
    from .nopdb import NoPdb


class Breakpoint(NoPdbContextManager):
    """A breakpoint that executes scheduled actions when hit.

    Breakpoints are typically created with :func:`nopdb.breakpoint`. The breakpoint
    object works as a context manager that removes the breakpoint on exit.
    """

    def __init__(
        self,
        nopdb: "NoPdb",
        scope: Scope,
        line: Optional[Union[int, str]] = None,
        cond: Optional[Union[str, bytes, CodeType]] = None,
    ):
        NoPdbContextManager.__init__(
            self, nopdb=nopdb, scope=scope, events=["call", "line"]
        )
        if line is None and scope.function is None:
            raise TypeError("A line must be given if no function is specified")
        if all(x is None for x in [scope.module, scope.file, scope.function]):
            raise TypeError("A module, file or function must be specified")

        self.scope = scope
        self.line = line
        self.cond = cond
        self._todo_list: List[Callable[[FrameType, str, Any], None]] = []

    def eval(
        self,
        expression: Union[str, bytes, CodeType],
        variables: Optional[Dict[str, Any]] = None,
    ) -> list:
        """Schedule an expression to be evaluated at the breakpoint.

        Args:
            expression (str, bytes or ~types.CodeType): A Python expression to be
                evaluated in the breakpoint's scope.
            variables (~typing.Dict[str, ~typing.Any], optional): External variables
                for the expression.

        Returns:
            list:
                An empty list that will later be filled with values of the expression.
        """
        results: list = []
        self._todo_list.append(
            functools.partial(
                self._do_eval,
                expression=expression,
                variables=variables,
                results=results,
            )
        )
        return results

    def exec(
        self,
        code: Union[str, bytes, CodeType],
        variables: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Schedule some code to be executed at the breakpoint.

        The code will be executed in the breakpoint's scope. Any changes to local
        variables (including newly defined variables) will be preserved in the local
        scope; note that this feature is somewhat experimental and may not work with
        Python implementations other than CPython and PyPy.

        Args:
            code (str, bytes or ~types.CodeType): Python source code to be executed in
                the breakpoint's scope.
            variables (~typing.Dict[str, ~typing.Any], optional): External variables
                for the code. These may not conflict with local variables and will
                *not* be preserved in the local scope.
        """
        self._todo_list.append(
            functools.partial(self._do_exec, code=code, variables=variables)
        )

    def debug(self, debugger_cls: Type[bdb.Bdb] = pdb.Pdb, **kwargs) -> None:
        """Schedule an interactive debugger to be entered at the breakpoint.

        Args:
            debugger_cls (~typing.Type[bdb.Bdb], optional): The debuger class. Defaults
                to :class:`pdb.Pdb`.
            **kwargs: Keyword arguments to pass to the debugger.
        """
        self._todo_list.append(
            functools.partial(self._do_debug, debugger_cls=debugger_cls, kwargs=kwargs)
        )

    @staticmethod
    def _do_eval(
        frame: FrameType,
        event: Union[str, bytes, CodeType],
        arg: Any,
        expression: str,
        results: list,
        variables: Optional[Dict[str, Any]],
    ) -> None:
        f_locals = {**frame.f_locals, **(variables or {})}
        result = eval(expression, dict(frame.f_globals), f_locals)
        results.append(result)

    @staticmethod
    def _do_exec(
        frame: FrameType,
        event: str,
        arg: Any,
        code: Union[str, bytes, CodeType],
        variables: Optional[Dict[str, Any]],
    ) -> None:
        if variables is None:
            variables = {}
        conflict_vars = frame.f_locals.keys() & variables.keys()
        if conflict_vars:
            raise RuntimeError(
                "The following external variables conflict with local ones: {}".format(
                    repr(conflict_vars).lstrip("{").rstrip("}")
                )
            )

        # Run the code with the external variables added, then remove them again
        f_locals = {**frame.f_locals, **variables}
        exec(code, frame.f_globals, f_locals)
        for name in list(f_locals.keys() & variables.keys()):
            del f_locals[name]

        # Update the frame
        for name in list(frame.f_locals.keys() - f_locals.keys()):
            del frame.f_locals[name]
        frame.f_locals.update(f_locals)
        update_locals(frame)

    @staticmethod
    def _do_debug(
        frame: FrameType,
        event: str,
        arg: Any,
        debugger_cls: Type[bdb.Bdb],
        kwargs: dict,
    ) -> None:
        debugger = get_nice_debugger(frame, debugger_cls, kwargs)
        sys.settrace(None)
        debugger.set_trace(frame=frame)
        debugger.trace_dispatch(frame, event, arg)

    def _callback(self, frame: FrameType, event: str, arg: Any) -> None:
        # If line is None, break when the function is called
        if self.line is None and event == "call":
            return
        # If line is an int, check the line number
        if isinstance(self.line, int) and frame.f_lineno != self.line:
            return
        # If line is a string, match the line text
        if isinstance(self.line, str):
            line_text = linecache.getline(
                frame.f_code.co_filename, frame.f_lineno, frame.f_globals
            )
            if self.line.strip() != line_text.strip():
                return
        # Evaluate condition if given
        if self.cond is not None and not eval(
            self.cond, frame.f_globals, frame.f_locals
        ):
            return

        for action in self._todo_list:
            action(frame, event, arg)


def get_update_locals():
    # fmt: off
    try:
        import ctypes
        locals_to_fast = ctypes.pythonapi.PyFrame_LocalsToFast

        def update_locals(frame: FrameType):
            locals_to_fast(ctypes.py_object(frame), ctypes.c_int(1))

        return update_locals
    except (ImportError, AttributeError):
        pass

    try:
        import __pypy__  # type: ignore
        return __pypy__.locals_to_fast
    except (ImportError, AttributeError):
        pass
    # fmt: on

    warnings.warn(
        "Unknown Python implementation (not CPython or PyPy). Local variable "
        "assignment probably will not work",
        RuntimeWarning,
    )

    def update_locals(frame: FrameType):
        pass

    return update_locals


update_locals = get_update_locals()



import inspect
from os import PathLike
import pathlib
from types import CodeType, FrameType, ModuleType, MethodType
from typing import Any, Callable, Optional, Tuple, Union


class Scope:
    def __init__(
        self,
        function: Optional[Union[Callable, str]] = None,
        module: Optional[ModuleType] = None,
        file: Optional[Union[str, PathLike]] = None,
        unwrap: bool = True,
    ):
        self.function = function
        self.module = module

        self._fn_name, self._fn_code, self._fn_self = None, None, None
        if isinstance(function, str):
            self._fn_name = function
        elif function is not None:
            self._fn_code, self._fn_self = _get_code_and_self(function, unwrap=unwrap)

        self.file: Optional[Union[str, pathlib.Path]] = None
        if isinstance(file, PathLike):
            self.file = pathlib.Path(file).resolve()
        else:
            self.file = file

    def match_frame(self, frame: FrameType) -> bool:
        if self._fn_code is not None:
            if frame.f_code is not self._fn_code:
                return False
            if self._fn_self is not None:
                # The function is a method; check if `self` is the correct object
                arg_info = inspect.getargvalues(frame)
                if len(arg_info.args) == 0:  # Just in case this somehow happens
                    return False
                if arg_info.locals[arg_info.args[0]] is not self._fn_self:
                    return False

        if self._fn_name is not None:
            if frame.f_code.co_name != self._fn_name:
                return False

        if self.module is not None:
            if frame.f_code.co_filename != self.module.__file__:
                return False

        if self.file is not None:
            file = frame.f_code.co_filename
            if file.endswith(">"):
                # Special filename: exact match
                if file != self.file:
                    return False
            elif isinstance(self.file, pathlib.Path):
                if pathlib.Path(file).resolve() != self.file:
                    return False
            elif not pathlib.Path(file).resolve().match(self.file):
                return False

        return True


def _get_code_and_self(fn: Callable, unwrap: bool) -> Tuple[CodeType, Any]:
    # First find the actual Python function that implements the callable, if possible.
    # Then if unwrap is True, try to follow the wrapper chain, assuming that the
    # wrappers are well-behaved (as opposed to, say, a Python function wrapping a
    # built-in). If not, we raise an error.

    def _unwrap(f):
        if unwrap:
            f = inspect.unwrap(f)
        return f

    exc: Optional[Exception] = None
    try:
        # Bound method
        if inspect.ismethod(fn):
            assert isinstance(fn, MethodType)
            return _unwrap(fn.__func__).__code__, fn.__self__
        # Regular function
        if inspect.isfunction(fn):
            return _unwrap(fn).__code__, None
        # Instance of a class that defines a __call__ method
        if hasattr(fn, "__class__") and hasattr(fn.__class__, "__call__"):
            return _unwrap(fn.__class__.__call__).__code__, fn
    except AttributeError as _exc:
        exc = _exc

    raise TypeError(
        "Could not find the code for {!r}. "
        "Please provide a pure Python callable".format(fn)
    ) from exc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nopdb']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'nopdb',
    'version': '0.1.0rc2',
    'description': 'Non-interactive Python debugger',
    'long_description': 'NoPdb: Non-interactive Python Debugger\n======================================\n|pypi-package| |docs-status| |test-status| |lint-status|\n\n* **Installation:** :code:`pip install nopdb`\n* **Docs:** https://nopdb.readthedocs.io/\n\nNoPdb is a **programmatic** (non-interactive) **debugger** for Python. This means it gives you access to\n**debugger-like superpowers** directly from your code. With NoPdb, you can:\n\n* **capture function calls**, including arguments, local variables, return values and stack traces\n* **set "breakpoints"** that trigger user-defined actions when hit, namely:\n\n  * **evaluate expressions** to retrieve their values later\n  * **execute arbitrary code**, including modifying local variables\n  * **enter an interactive debugger** like `pdb`\n\nNoPdb is also a convenient tool for inspecting **machine learning model internals**. For example, `this notebook <https://colab.research.google.com/github/cifkao/nopdb/blob/main/docs/pytorch_tutorial.ipynb>`_ shows how to\nuse it to visualize Transformer attention in PyTorch.\n\nNoPdb should run at least under CPython and PyPy. Most features should work under any implementation\nas long as it has :code:`sys.settrace()`.\n\n**Note:** This project is in its early development stage. Contributions and improvement ideas are welcome.\n\nCapturing function calls\n------------------------\n\nThe functions :code:`capture_call()` and :code:`capture_calls()` allow\ncapturing useful information about calls to a given function.\nThey are typically used as context managers, e.g.:\n\n.. code-block:: python\n\n    with nopdb.capture_calls(fn) as calls:\n        some_code_that_calls_fn()\n\n        print(calls)  # see details about how fn() was called\n\nThe information we can retrieve includes the function\'s arguments, return value, local variables and stack trace. For example:\n\n.. code-block:: python\n\n    >>> with nopdb.capture_call(f) as call:\n    ...     g(1)\n    >>> call\n    CallCapture(name=\'f\', args=OrderedDict(x=1, y=1), return_value=4)\n    >>> call.print_stack()\n      File "<stdin>", line 2, in <module>\n      File "<stdin>", line 2, in g\n      File "<stdin>", line 1, in f\n    >>> call.args[\'x\']\n    1\n    >>> call.return_value\n    4\n    >>> call.locals\n    {\'y\': 1, \'x\': 1, \'z\': 2}\n\nSetting breakpoints\n-------------------\n\nLike conventional debuggers, NoPdb can set breakpoints. However, because NoPdb is a\n*non-interactive* debugger, its breakpoints do not actually stop the execution of the\nprogram. Instead, they allow executing actions scheduled in advance, such as\nevaluating expressions.\n\nTo set a breakpoint, call the :code:`breakpoint()` function. A breakpoint object\nis returned, allowing to schedule actions using its methods such as\n:code:`eval()` and :code:`exec()`. For example:\n\n.. code-block:: python\n\n   # Break at line 3 of the file or notebook cell where f is defined\n   with nopdb.breakpoint(function=f, line=3) as bp:\n       x = bp.eval("x")             # Schedule an expression\n       type_y = bp.eval("type(y)")  # Another one\n       bp.exec("print(y)")          # Schedule a print statement\n\n       some_code_that_calls_f()\n\n   print(x, type_y)  # Retrieve the captured values\n\nThere are other ways to specify the breakpoint location. For example:\n\n.. code-block:: python\n\n   # Break at any line with the given source code in the given file\n   with nopdb.breakpoint(file="pathlib.py", line="return obj") as bp:\n       ...\n\n   # Break as soon as any function with the given name is called\n   with nopdb.breakpoint(function="myfunc") as bp:\n       ...\n\nNot only can we capture values, we can also modify them!\n\n.. code-block:: python\n\n    >>> with nopdb.breakpoint(function=f, line=3) as bp:\n    ...     # Get the value of x, then increment it, then get the new value\n    ...     x_before = bp.eval(\'x\')\n    ...     bp.exec(\'x += 1\')\n    ...     x_after = bp.eval(\'x\')\n    ...\n    ...     some_code_that_calls_f()\n    >>> x_before\n    [2]\n    >>> x_after\n    [3]\n\n.. |pypi-package| image:: https://badge.fury.io/py/nopdb.svg?\n   :target: https://pypi.org/project/nopdb/\n   :alt: PyPI Package\n.. |docs-status| image:: https://readthedocs.org/projects/nopdb/badge/?version=latest\n   :target: https://nopdb.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n.. |test-status| image:: https://github.com/cifkao/nopdb/actions/workflows/test.yml/badge.svg\n   :target: https://github.com/cifkao/nopdb/actions/workflows/test.yml\n   :alt: Lint Status\n.. |lint-status| image:: https://github.com/cifkao/nopdb/actions/workflows/lint.yml/badge.svg\n   :target: https://github.com/cifkao/nopdb/actions/workflows/lint.yml\n   :alt: Lint Status\n',
    'author': 'Ondřej Cífka',
    'author_email': 'cifkao@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cifkao/nopdb',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
